@@ -1,4 +1,4 @@
-# Copyright 2018 Hugh Salimbeni, Artem Artemev @awav
+# Copyright 2018-2020 Hugh Salimbeni, Artem Artemev @awav, ST John @st--
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,23 +14,31 @@
 
 import abc
 import functools
-from typing import Callable, List, Union, Tuple
+from typing import Callable, Optional, Sequence, Tuple, Union
 
 import numpy as np
 import tensorflow as tf
 
-from ..base import Parameter
+from ..base import Parameter, _to_constrained
 
 Scalar = Union[float, tf.Tensor, np.ndarray]
+LossClosure = Callable[[], tf.Tensor]
+NatGradParameters = Union[Tuple[Parameter, Parameter], Tuple[Parameter, Parameter, "XiTransform"]]
 
 __all__ = [
     "NaturalGradient",
     "XiTransform",
-    "XiSqrtMeanVar",
     "XiNat",
+    "XiSqrtMeanVar",
 ]
 
 
+#
+# Xi transformations necessary for natural gradient optimizer.
+# Abstract class and two implementations: XiNat and XiSqrtMeanVar.
+#
+
+
 class XiTransform(metaclass=abc.ABCMeta):
     """
     XiTransform is the base class that implements three transformations necessary
@@ -39,8 +47,9 @@ class XiTransform(metaclass=abc.ABCMeta):
     the parameters pairs are always of shape (N, D) and (D, N, N).
     """
 
+    @staticmethod
     @abc.abstractmethod
-    def meanvarsqrt_to_xi(self, mean, varsqrt):
+    def meanvarsqrt_to_xi(mean, varsqrt):
         """
         Transforms the parameter `mean` and `varsqrt` to `xi1`, `xi2`
 
@@ -48,10 +57,10 @@ class XiTransform(metaclass=abc.ABCMeta):
         :param varsqrt: the varsqrt parameter (D, N, N)
         :return: tuple (xi1, xi2), the xi parameters (N, D), (D, N, N)
         """
-        pass  # pragma: no cover
 
+    @staticmethod
     @abc.abstractmethod
-    def xi_to_meanvarsqrt(self, xi1, xi2):
+    def xi_to_meanvarsqrt(xi1, xi2):
         """
         Transforms the parameter `xi1`, `xi2` to `mean`, `varsqrt`
 
@@ -59,10 +68,10 @@ class XiTransform(metaclass=abc.ABCMeta):
         :param xi2: the ξ₂ parameter
         :return: tuple (mean, varsqrt), the meanvarsqrt parameters
         """
-        pass  # pragma: no cover
 
+    @staticmethod
     @abc.abstractmethod
-    def naturals_to_xi(self, nat1, nat2):
+    def naturals_to_xi(nat1, nat2):
         """
         Applies the transform so that `nat1`, `nat2` is mapped to `xi1`, `xi2`
 
@@ -70,83 +79,191 @@ class XiTransform(metaclass=abc.ABCMeta):
         :param nat2: the θ₂ parameter
         :return: tuple `xi1`, `xi2`
         """
-        pass  # pragma: no cover
+
+
+class XiNat(XiTransform):
+    """
+    This is the default transform. Using the natural directly saves the forward mode
+    gradient, and also gives the analytic optimal solution for gamma=1 in the case
+    of Gaussian likelihood.
+    """
+
+    @staticmethod
+    def meanvarsqrt_to_xi(mean, varsqrt):
+        return meanvarsqrt_to_natural(mean, varsqrt)
+
+    @staticmethod
+    def xi_to_meanvarsqrt(xi1, xi2):
+        return natural_to_meanvarsqrt(xi1, xi2)
+
+    @staticmethod
+    def naturals_to_xi(nat1, nat2):
+        return nat1, nat2
+
+
+class XiSqrtMeanVar(XiTransform):
+    """
+    This transformation will perform natural gradient descent on the model parameters,
+    so saves the conversion to and from Xi.
+    """
+
+    @staticmethod
+    def meanvarsqrt_to_xi(mean, varsqrt):
+        return mean, varsqrt
+
+    @staticmethod
+    def xi_to_meanvarsqrt(xi1, xi2):
+        return xi1, xi2
+
+    @staticmethod
+    def naturals_to_xi(nat1, nat2):
+        return natural_to_meanvarsqrt(nat1, nat2)
 
 
 class NaturalGradient(tf.optimizers.Optimizer):
-    def __init__(self, gamma: Scalar, name=None):
+    """
+    Implements a natural gradient descent optimizer for variational models
+    that are based on a distribution q(u) = N(q_mu, q_sqrt q_sqrtᵀ) that is
+    parameterized by mean q_mu and lower-triangular Cholesky factor q_sqrt
+    of the covariance.
+
+    Note that this optimizer does not implement the standard API of
+    tf.optimizers.Optimizer. Its only public method is minimize(), which has
+    a custom signature (var_list needs to be a list of (q_mu, q_sqrt) tuples,
+    where q_mu and q_sqrt are gpflow.Parameter instances, not tf.Variable).
+
+    When using in your work, please cite
+
+        @inproceedings{salimbeni18,
+            title={Natural Gradients in Practice: Non-Conjugate Variational Inference in Gaussian Process Models},
+            author={Salimbeni, Hugh and Eleftheriadis, Stefanos and Hensman, James},
+            booktitle={AISTATS},
+            year={2018}
+    """
+
+    def __init__(self, gamma: Scalar, xi_transform: XiTransform = XiNat(), name=None):
+        """
+        :param gamma: natgrad step length
+        :param xi_transform: default ξ transform (can be overridden in the call to minimize())
+            The XiNat default choice works well in general.
+        """
         name = self.__class__.__name__ if name is None else name
         super().__init__(name)
         self.gamma = gamma
+        self.xi_transform = xi_transform
 
     def minimize(
-        self,
-        loss_fn: Callable,
-        var_list: List[
-            Union[Tuple[Parameter, Parameter], Tuple[Parameter, Parameter, XiTransform]]
-        ],
+        self, loss_fn: LossClosure, var_list: Sequence[NatGradParameters],
     ):
         """
         Minimizes objective function of the model.
         Natural Gradient optimizer works with variational parameters only.
-        There are two supported ways of transformation for parameters:
-            - XiNat
-            - XiSqrtMeanVar
-        Custom transformations are also possible, they should implement
-        `XiTransform` interface.
 
-            :param loss_fn: Loss function.
-            :param var_list: List of pair tuples of variational parameters or
-                triplet tuple with variational parameters and ξ transformation.
-                By default, all parameters goes through XiNat() transformation.
-                For example your `var_list` can look as,
-                ```
-                var_list = [
-                    (q_mu1, q_sqrt1),
-                    (q_mu2, q_sqrt2, XiSqrtMeanVar())
-                ]
-                ```
+        :param loss_fn: Loss function.
+        :param var_list: List of pair tuples of variational parameters or
+            triplet tuple with variational parameters and ξ transformation.
+            If ξ is not specified, will use self.xi_transform.
+            For example, `var_list` could be
+            ```
+            var_list = [
+                (q_mu1, q_sqrt1),
+                (q_mu2, q_sqrt2, XiSqrtMeanVar())
+            ]
+            ```
+
+        GPflow implements the `XiNat` (default) and `XiSqrtMeanVar` transformations
+        for parameters. Custom transformations that implement the `XiTransform`
+        interface are also possible.
         """
-        parameters = [(v[0], v[1], v[2] if len(v) > 2 else XiNat()) for v in var_list]
+        parameters = [(v[0], v[1], (v[2] if len(v) > 2 else None)) for v in var_list]
         self._natgrad_steps(loss_fn, parameters)
 
     def _natgrad_steps(
-        self, loss_fn: Callable, parameters: List[Tuple[Parameter, Parameter, XiTransform]]
-    ):
-        def natural_gradient_step(q_mu, q_sqrt, xi_transform):
-            self._natgrad_step(loss_fn, q_mu, q_sqrt, xi_transform)
-
-        with tf.name_scope(f"{self._name}/natural_gradient_steps"):
-            list(map(natural_gradient_step, *zip(*parameters)))
-
-    def _natgrad_step(
-        self, loss_fn: Callable, q_mu: Parameter, q_sqrt: Parameter, xi_transform: XiTransform
+        self,
+        loss_fn: LossClosure,
+        parameters: Sequence[Tuple[Parameter, Parameter, Optional[XiTransform]]],
     ):
         """
+        Computes gradients of loss_fn() w.r.t. q_mu and q_sqrt, and updates
+        these parameters using the natgrad backwards step, for all sets of
+        variational parameters passed in.
+
+        :param loss_fn: Loss function.
+        :param parameters: List of tuples (q_mu, q_sqrt, xi_transform)
+        """
+        q_mus, q_sqrts, xis = zip(*parameters)
+        unconstrained_variables = [
+            p.unconstrained_variable for params in (q_mus, q_sqrts) for p in params
+        ]
+
+        with tf.GradientTape(watch_accessed_variables=False) as tape:
+            tape.watch(unconstrained_variables)
+            loss = loss_fn()
+
+        q_mu_grads, q_sqrt_grads = tape.gradient(loss, [q_mus, q_sqrts])
+        # NOTE that these are the gradients in *unconstrained* space
+
+        with tf.name_scope(f"{self._name}/natural_gradient_steps"):
+            for q_mu_grad, q_sqrt_grad, q_mu, q_sqrt, xi_transform in zip(
+                q_mu_grads, q_sqrt_grads, q_mus, q_sqrts, xis
+            ):
+                self._natgrad_apply_gradients(q_mu_grad, q_sqrt_grad, q_mu, q_sqrt, xi_transform)
+
+    def _natgrad_apply_gradients(
+        self,
+        q_mu_grad: tf.Tensor,
+        q_sqrt_grad: tf.Tensor,
+        q_mu: Parameter,
+        q_sqrt: Parameter,
+        xi_transform: Optional[XiTransform] = None,
+    ):
+        """
+        This function does the backward step on the q_mu and q_sqrt parameters,
+        given the gradients of the loss function with respect to their unconstrained
+        variables. I.e., it expects the arguments to come from
+
+            with tf.GradientTape() as tape:
+                loss = loss_function()
+            q_mu_grad, q_mu_sqrt = tape.gradient(loss, [q_mu, q_sqrt])
+
+        (Note that tape.gradient() returns the gradients in *unconstrained* space!)
+
         Implements equation [10] from
 
         @inproceedings{salimbeni18,
-            title={Natural Gradients in Practice: Non-Conjugate  Variational Inference in Gaussian Process Models},
+            title={Natural Gradients in Practice: Non-Conjugate Variational Inference in Gaussian Process Models},
             author={Salimbeni, Hugh and Eleftheriadis, Stefanos and Hensman, James},
             booktitle={AISTATS},
             year={2018}
 
         In addition, for convenience with the rest of GPflow, this code computes ∂L/∂η using
-        the chain rule:
+        the chain rule (the following assumes a numerator layout where the gradient is a row
+        vector; note that TensorFlow actually returns a column vector), where L is the loss:
 
-        ∂L/∂η = (∂L / ∂[q_μ, q_sqrt])(∂[q_μ, q_sqrt] / ∂η)
+        ∂L/∂η = (∂L / ∂[q_mu, q_sqrt])(∂[q_mu, q_sqrt] / ∂η)
 
         In total there are three derivative calculations:
-        natgrad L w.r.t ξ  = (∂ξ / ∂θ) [(∂L / ∂[q_μ, q_sqrt]) (∂[q_μ, q_sqrt] / ∂η)]^T
+        natgrad of L w.r.t ξ  = (∂ξ / ∂θ) [(∂L / ∂[q_mu, q_sqrt]) (∂[q_mu, q_sqrt] / ∂η)]ᵀ
 
-        Note that if ξ = θ or [q_μ, q_sqrt] some of these calculations are the identity.
+        Note that if ξ = θ (i.e. [q_mu, q_sqrt]) some of these calculations are the identity.
         In the code η = eta, ξ = xi, θ = nat.
+
+        :param q_mu_grad: gradient of loss w.r.t. q_mu (in unconstrained space)
+        :param q_sqrt_grad: gradient of loss w.r.t. q_sqrt (in unconstrained space)
+        :param q_mu: parameter for the mean of q(u)
+        :param q_sqrt: parameter for the square root of the covariance of q(u)
+        :param xi_transform: the ξ transform to use (self.xi_transform if not specified)
         """
+        if xi_transform is None:
+            xi_transform = self.xi_transform
+
+        # 1) the ordinary gpflow gradient
+        dL_dmean = _to_constrained(q_mu_grad, q_mu.transform)
+        dL_dvarsqrt = _to_constrained(q_sqrt_grad, q_sqrt.transform)
 
         with tf.GradientTape(persistent=True, watch_accessed_variables=False) as tape:
             tape.watch([q_mu.unconstrained_variable, q_sqrt.unconstrained_variable])
 
-            loss = loss_fn()
             # the three parameterizations as functions of [q_mu, q_sqrt]
             eta1, eta2 = meanvarsqrt_to_expectation(q_mu, q_sqrt)
             # we need these to calculate the relevant gradients
@@ -162,10 +279,6 @@ class NaturalGradient(tf.optimizers.Optimizer):
                         [xi1_nat, xi2_nat], [nat1, nat2], output_gradients=dummy_tensors
                     )
 
-        # 1) the ordinary gpflow gradient
-        dL_dmean, dL_dvarsqrt = tape.gradient(loss, [q_mu, q_sqrt])
-        dL_dvarsqrt = q_sqrt.transform.forward(dL_dvarsqrt)
-
         # 2) the chain rule to get ∂L/∂η, where η (eta) are the expectation parameters
         dL_deta1, dL_deta2 = tape.gradient(
             meanvarsqrt, [eta1, eta2], output_gradients=[dL_dmean, dL_dvarsqrt]
@@ -184,7 +297,7 @@ class NaturalGradient(tf.optimizers.Optimizer):
         xi1_new = xi1 - self.gamma * nat_dL_xi1
         xi2_new = xi2 - self.gamma * nat_dL_xi2
 
-        # Transform back to the model parameters [q_μ, q_sqrt]
+        # Transform back to the model parameters [q_mu, q_sqrt]
         mean_new, varsqrt_new = xi_transform.xi_to_meanvarsqrt(xi1_new, xi2_new)
 
         q_mu.assign(mean_new)
@@ -192,51 +305,10 @@ class NaturalGradient(tf.optimizers.Optimizer):
 
     def get_config(self):
         config = super().get_config()
-        config.update(
-            {"gamma": self._serialize_hyperparameter("gamma"),}
-        )
+        config.update({"gamma": self._serialize_hyperparameter("gamma")})
         return config
 
 
-#
-# Xi transformations necessary for natural gradient optimizer.
-# Abstract class and two implementations: XiNat and XiSqrtMeanVar.
-#
-
-
-class XiNat(XiTransform):
-    """
-    This is the default transform. Using the natural directly saves the forward mode
-     gradient, and also gives the analytic optimal solution for gamma=1 in the case
-     of Gaussian likelihood.
-    """
-
-    def meanvarsqrt_to_xi(self, mean, varsqrt):
-        return meanvarsqrt_to_natural(mean, varsqrt)
-
-    def xi_to_meanvarsqrt(self, xi1, xi2):
-        return natural_to_meanvarsqrt(xi1, xi2)
-
-    def naturals_to_xi(self, nat1, nat2):
-        return nat1, nat2
-
-
-class XiSqrtMeanVar(XiTransform):
-    """
-    This transformation will perform natural gradient descent on the model parameters,
-    so saves the conversion to and from Xi.
-    """
-
-    def meanvarsqrt_to_xi(self, mean, varsqrt):
-        return mean, varsqrt
-
-    def xi_to_meanvarsqrt(self, xi1, xi2):
-        return xi1, xi2
-
-    def naturals_to_xi(self, nat1, nat2):
-        return natural_to_meanvarsqrt(nat1, nat2)
-
-
 #
 # Auxiliary gaussian parameter conversion functions.
 #
